#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/xpm.h>
#include <X11/extensions/shape.h>
#include <time.h>
#include <fcntl.h>
#include <X11/Xatom.h>

#define DUAL

#ifdef DUAL
#include "frm2.xpm"
#include "mask2.xbm"
#else
#include "frm1.xpm"
#include "mask1.xbm"
#endif
#include "led.xpm"

#define VERSION "0.8"

#define PROC_STAT "/proc/stat"
#define PROC_CPUINFO "/proc/cpuinfo"

int ICONIFIED=0; /* default is not iconified */
int bogo=0;
#define CPU_H 15
#define CPU_W 54
/* with shape */
int posx[3] = {5, 5, 5};
int posy[3] = {5, 5+19, 5+19*2};
#define LED_W 8
#define LED_H 9
int ledx[] = {0, 7, 14, 21, 28, 35, 42, 49, 56, 63, 70};
int ledy[] = {0, 0,  0,  0,  0,  0,  0,  0,  0,  0,  0};
#ifdef DUAL
#define LED_BOX 2
#else
#define LED_BOX 1
#endif

/* X11 Variablen *****************************************************/
Display *dpy;		/* welches DISPLAY */
Window Root;			/* Hintergrund-Drawable */
int screen;
int x_fd;
int d_depth;
XSizeHints mysizehints;
XWMHints mywmhints;
Pixel back_pix, fore_pix;
GC NormalGC;
Window iconwin, win;			 /* My home is my window */
char *ProgName;
char *Geometry;
char Execute[] = "echo no program has been specified >/dev/console";
char *ERR_colorcells = "not enough free color cells\n";
char *ampers = " &";
/* XPM Variablen *****************************************************/
typedef struct _XpmIcon {
		Pixmap pixmap;
		Pixmap mask;
		XpmAttributes attributes;
}				XpmIcon;

XpmIcon wmdual, led;
XpmIcon visible;
time_t actualtime;
long actualmin;

/* lokale Funktionen *************************************************/
#define MW_EVENTS	 (ExposureMask | ButtonPressMask | StructureNotifyMask)
#define FALSE 0
void GetXPM(void);
static void CreatePixmap(Display *dpy, Window Root, char **data, XpmIcon* icon);
Pixel GetColor(char *name);
void RedrawWindow( XpmIcon *v);
void InitLoad();
void InsertLoad();
/*********************************************************************/
/*********************************************************************/
int updatespeed = 4;
static char *help_message[] = {
"where options include:",
"  -u <secs>              : updatespeed",
"  -b                     : pritn bogoMIPS",       
"  -exe <program>         : program to start on click",
"  -position [+|-]x[+|-]y : position of wmdual",
"  -iconic                : start up as icon",
NULL
};

void usage()
{
	char **cpp;

	fprintf(stderr,"wmdual ver %s\n",VERSION);
	fprintf(stderr,"usage:	%s [-options ...] \n", ProgName);
	for (cpp = help_message; *cpp; cpp++) {
		fprintf(stderr, "%s\n", *cpp);
	}
	fprintf(stderr,"\n");
	exit(1);
}
int main(int argc,char* argv[])
{
	int i;
	unsigned int borderwidth ;
	char *display_name = NULL; 
	char *wname = "wmdual";
	XGCValues gcv;
	unsigned long gcm;
	XEvent Event;
	XTextProperty name;
	XClassHint classHint;
	Pixmap pixmask;
	ProgName = argv[0];
	Geometry = "";
	
	/* Parse command line options */
	ProgName = argv[0];

	for(i=1;i<argc;i++) {
		char *arg= argv[i];

		if (arg[0] == '-') {
			switch(arg[1]) {
			case 'u':
				if (++i >= argc) usage();
				sscanf(argv[i],"%d",&updatespeed);
				continue;
			case 'e':
				if(++i >=argc) usage();
				strcpy(&Execute[0], argv[i]);
				strcat(&Execute[0], " &");
				continue;
			case 'p':
				if(++i >=argc) usage();
				Geometry = argv[i];
				continue;
			case 'i':
				ICONIFIED=1;
				continue;
			case 'b':
				bogo=1;
				continue;
			default:
				usage();
			}
		}
	}

	/* Open the display */
	if (!(dpy = XOpenDisplay(display_name)))	
		{ 
			fprintf(stderr,"wmdual: can't open display %s\n", 
				XDisplayName(display_name)); 
			exit (1); 
		} 
	screen= DefaultScreen(dpy);
	Root = RootWindow(dpy, screen);
	d_depth = DefaultDepth(dpy, screen);
	x_fd = XConnectionNumber(dpy);
	
	/* Icon Daten nach XImage konvertieren */
	GetXPM();
	
	/* Create a window to hold the banner */
	mysizehints.flags= USSize|USPosition;
	mysizehints.x = 0;
	mysizehints.y = 0;

	back_pix = GetColor("white");
	fore_pix = GetColor("black");

	XWMGeometry(dpy, screen, Geometry, NULL, (borderwidth =1), &mysizehints,
				&mysizehints.x,&mysizehints.y,&mysizehints.width,&mysizehints.height, &i); 

	mysizehints.width = wmdual.attributes.width;
	mysizehints.height= wmdual.attributes.height;

	win = XCreateSimpleWindow(dpy,Root,mysizehints.x,mysizehints.y,
					mysizehints.width,mysizehints.height,
					borderwidth,fore_pix,back_pix);
	iconwin = XCreateSimpleWindow(dpy,win,mysizehints.x,mysizehints.y,
				mysizehints.width,mysizehints.height,
				borderwidth,fore_pix,back_pix);



	/* Hints aktivieren */
	XSetWMNormalHints(dpy, win, &mysizehints);
	classHint.res_name =	"wmdual";
	classHint.res_class = "WMDUAL";
	XSetClassHint(dpy, win, &classHint);

	XSelectInput(dpy,win,MW_EVENTS);
	XSelectInput(dpy,iconwin,MW_EVENTS);
	
	if (XStringListToTextProperty(&wname, 1, &name) ==0) {
		fprintf(stderr, "wmdual: can't allocate window name\n");
		exit(-1);
	}
	XSetWMName(dpy, win, &name);
	
	/* Create a GC for drawing */
	gcm = GCForeground|GCBackground|GCGraphicsExposures;
	gcv.foreground = fore_pix;
	gcv.background = back_pix;
	gcv.graphics_exposures = FALSE;
	NormalGC = XCreateGC(dpy, Root, gcm, &gcv);	

		pixmask = XCreateBitmapFromData(dpy, win, mask_bits, mask_width, 
						mask_height);
		XShapeCombineMask(dpy, win, ShapeBounding, 0, 0, pixmask, ShapeSet);
		XShapeCombineMask(dpy, iconwin, ShapeBounding, 0, 0, pixmask, ShapeSet);
	
	mywmhints.initial_state = WithdrawnState;
	mywmhints.icon_window = iconwin;
	mywmhints.icon_x = mysizehints.x;
	mywmhints.icon_y = mysizehints.y;
	mywmhints.window_group = win;
	mywmhints.flags = StateHint | IconWindowHint | IconPositionHint 
			| WindowGroupHint;
	XSetWMHints(dpy, win, &mywmhints); 

	XSetCommand(dpy, win, argv, argc);
	XMapWindow(dpy,win);

	XCopyArea(dpy,visible.pixmap,wmdual.pixmap,NormalGC,
		posx[0],posy[0],CPU_W,CPU_H,posx[0],posy[0]);
	XCopyArea(dpy,wmdual.pixmap,visible.pixmap,NormalGC,
		0,0,mysizehints.width,mysizehints.height,0,0);
	XCopyArea(dpy,visible.pixmap,visible.pixmap,NormalGC,
		posx[0],posy[1],CPU_W,CPU_H,posx[0],posy[0]);
	InitLoad();
/*	InsertLoad(); */
	RedrawWindow(&visible);
	while(1) {
			if (actualtime != time(0)) {
		actualtime = time(0);
		if(actualtime % updatespeed == 0) {
				InsertLoad();
		}
		RedrawWindow(&visible);

	}
			
			/* read a packet */
			while (XPending(dpy)) {
		XNextEvent(dpy,&Event);
		switch(Event.type) {
			case Expose:
				if(Event.xexpose.count == 0 )
		RedrawWindow(&visible);
				break;
			case ButtonPress:
				system(Execute);
				break;
			case DestroyNotify:
		/*
		 * Don't need these as XCloseDisplay() will
		 * free them. 
							XFreeGC(dpy, NormalGC);
							XDestroyWindow(dpy, win);
				XDestroyWindow(dpy, iconwin);
		 */
							XCloseDisplay(dpy);
				exit(0); 
			default:
				break;			
			}
	}
			XFlush(dpy);
#ifdef SYSV
			poll((struct poll *) 0, (size_t) 0, 50);
#else
			usleep(50000L);			/* 5/100 sec */
#endif
		}
	return 0;
}
/****************************************************************************/
void nocolor(char *a, char *b)
{
 fprintf(stderr,"wmdual: can't %s %s\n", a,b);
}
/****************************************************************************/
/* Konvertiere XPMIcons nach XImage */
void GetXPM(void)
{
	XColor col;
	XWindowAttributes attributes;
	char led1[22];
	char led2[22];
	int ret;

	/* for the colormap */
	XGetWindowAttributes(dpy,Root,&attributes);

	CreatePixmap(dpy, Root, frm_xpm, &wmdual);
	CreatePixmap(dpy, Root, frm_xpm, &visible);
	CreatePixmap(dpy, Root, led_xpm, &led);
}
/****************************************************************************/
static void CreatePixmap(Display *dpy, Window Root, char **data, XpmIcon* icon)
{
	int ret;

	icon->attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions
				 | XpmExactColors | XpmCloseness);
	icon->attributes.exactColors=False;
	icon->attributes.closeness=40000;
	ret = XpmCreatePixmapFromData(dpy, Root, data, &(icon->pixmap), 
				&(icon->mask), &(icon->attributes));
	if(ret != XpmSuccess)
		{fprintf(stderr, ERR_colorcells);exit(1);}
}
/****************************************************************************/
/* Removes expose events for a specific window from the queue */
int flush_expose (Window w)
{
	XEvent dummy;
	int i=0;
	
	while (XCheckTypedWindowEvent (dpy, w, Expose, &dummy))i++;
	return i;
}

/****************************************************************************/
/* Draws the icon window */
void RedrawWindow( XpmIcon *v)
{
	flush_expose (iconwin);
	XCopyArea(dpy,v->pixmap,iconwin,NormalGC,
			0,0,v->attributes.width, v->attributes.height,0,0);
	flush_expose (win);
	XCopyArea(dpy,v->pixmap,win,NormalGC,
			0,0,v->attributes.width, v->attributes.height,0,0);

}
/****************************************************************************/
Pixel GetColor(char *name)
{
	XColor color;
	XWindowAttributes attributes;

	XGetWindowAttributes(dpy,Root,&attributes);
	color.pixel = 0;
	 if (!XParseColor (dpy, attributes.colormap, name, &color)) 
		 {
			 nocolor("parse",name);
		 }
	 else if(!XAllocColor (dpy, attributes.colormap, &color)) 
		 {
			 nocolor("alloc",name);
		 }
	return color.pixel;
}
/*********************************************************************/
#define MAX_CPU 2
static unsigned long cpu[MAX_CPU+1][4];
static int max_cpu=-1;
void InitLoad()
{
	unsigned int cpu_clock = 0;
	double bogomips = 0.0;
	char buf[256];
 	int fd,len;
	unsigned long n[4];
	int id;
	int i,j;
	FILE* fp;

	fp = fopen(PROC_STAT,"r");
	while (fgets(buf,sizeof(buf),fp)) {
		if (strncmp(buf,"cpu ",4) == 0) {
			sscanf(buf,"cpu %u %u %u %u",&n[0],&n[1],&n[2],&n[3]);
			cpu[0][0] = n[0];
			cpu[0][1] = n[1];
			cpu[0][2] = n[2];
			cpu[0][3] = n[3];
		} else if (strncmp(buf,"cpu",3) == 0) {
			sscanf(buf,"cpu%d %u %u %u %u",&id,&n[0],&n[1],&n[2],&n[3]);
			if (MAX_CPU > id) {
				cpu[id+1][0] = n[0];
				cpu[id+1][1] = n[1];
				cpu[id+1][2] = n[2];
				cpu[id+1][3] = n[3];
				if (id > max_cpu)
					max_cpu = id;
			}
		}
	}
	fclose(fp);
	if (!bogo) {
		double tmp=0.0;
		fp = fopen(PROC_CPUINFO,"r");
		while (fgets(buf,sizeof(buf),fp)) {
			if (strncmp(buf,"cpu MHz",7) == 0) {
				sscanf(buf,"cpu MHz				 : %lf",&tmp);
				break;
			}
		}
		cpu_clock = (unsigned int)tmp;
		fclose(fp);
	}
#ifdef DEBUG
	fprintf(stderr,"clock %u MHz\n",cpu_clock);
#endif
	if (cpu_clock > 0) {
		XCopyArea(dpy,led.pixmap,visible.pixmap,NormalGC,
			28,9,21,11,posx[LED_BOX]+32,posy[LED_BOX]+4);
		sprintf(buf,"%u",cpu_clock);
		len = 4-strlen(buf);
		for (i = 0;i < strlen(buf);i++,len++) {
			j = buf[i] - '0';
			XCopyArea(dpy,led.pixmap,visible.pixmap,NormalGC,
				ledx[j],ledy[j],LED_W,LED_H,posx[LED_BOX]+1+len*7,posy[LED_BOX]+4);
		}
	} else {
		double tmp=0.0;
		fp = fopen(PROC_CPUINFO,"r");
		while (fgets(buf,sizeof(buf),fp)) {
			if (strncmp(buf,"bogomips",7) == 0) {
				sscanf(buf,"bogomips				: %lf",&tmp);
				bogomips += tmp;
			}
		}
		fclose(fp);
#ifdef DEBUG
		fprintf(stderr,"bogomips %lf\n",bogomips);
#endif
		if (bogomips > 0) {
			XCopyArea(dpy,led.pixmap,visible.pixmap,NormalGC,
				0,9,24,11,posx[LED_BOX]+29,posy[LED_BOX]+4);
			sprintf(buf,"%u",(unsigned int)bogomips);
			len = 4-strlen(buf);
			for (i = 0;i < strlen(buf);i++,len++) {
				j = buf[i] - '0';
				XCopyArea(dpy,led.pixmap,visible.pixmap,NormalGC,
					ledx[j],ledy[j],LED_W,LED_H,posx[LED_BOX]+1+len*7,posy[LED_BOX]+4);
			}
		}
	}
}
void InsertLoad()
{
	char buf[256];
	FILE* fp;
	int id;
	int i;
	unsigned long n[4],total;
	int free,used;
	unsigned int max = CPU_W;

	fp = fopen(PROC_STAT,"r");
	while (fgets(buf,sizeof(buf),fp)) {
		if (strncmp(buf,"cpu ",4) == 0) {
			if (max_cpu > 0)
				continue;
			sscanf(buf,"cpu %u %u %u %u",&n[0],&n[1],&n[2],&n[3]);
			used = n[0] - cpu[0][0]+ n[1] - cpu[0][1] + n[2] - cpu[0][2];
			free = n[3] - cpu[0][3];
			total = used + free;
			cpu[0][0] = n[0];
			cpu[0][1] = n[1];
			cpu[0][2] = n[2];
			cpu[0][3] = n[3];
			used = total ? (max*(double)used/total+0.5) : 0;
#ifdef DEBUG
			fprintf(stderr,"cpu used %d\n",used);
#endif
			XCopyArea(dpy,wmdual.pixmap,visible.pixmap,NormalGC,
				posx[0],posy[0],used,CPU_H,posx[0],posy[0]);
#ifdef DUAL
			XCopyArea(dpy,wmdual.pixmap,visible.pixmap,NormalGC,
				posx[1],posy[1],max-used,CPU_H,posx[0]+used,posy[0]);
#endif

		} else if (strncmp(buf,"cpu",3) == 0) { /* cpu0..cpu1 cpuN */
			sscanf(buf,"cpu%d %u %u %u %u",&id,&n[0],&n[1],&n[2],&n[3]);
			if (MAX_CPU > id) {
				used = n[0]-cpu[id+1][0]+n[1]-cpu[id+1][1]+n[2]-cpu[id+1][2];
				free = n[3] - cpu[id+1][3];
				total = used + free;
				cpu[id+1][0] = n[0];
				cpu[id+1][1] = n[1];
				cpu[id+1][2] = n[2];
				cpu[id+1][3] = n[3];
				used = total ? (max*(double)used/total+0.5) : 0;
#ifdef DEBUG
				fprintf(stderr,"cpu%d used %d\n",id,used);
#endif
				XCopyArea(dpy,wmdual.pixmap,visible.pixmap,NormalGC,
					posx[0],posy[0],used,CPU_H,posx[0],posy[id]);
#ifdef DUAL
				XCopyArea(dpy,wmdual.pixmap,visible.pixmap,NormalGC,
					posx[1],posy[1],max-used,CPU_H,posx[0]+used,posy[id]);
#endif
			}
		}
	}
	fclose(fp);
}

